/*
 * This file is part of Deliantra, the Roguelike Realtime MMORPG.
 *
 * Copyright (©) 2005,2006,2007,2008,2009,2010,2011,2012 Marc Alexander Lehmann / Robin Redeker / the Deliantra team
 * Copyright (©) 2001 Mark Wedel & Crossfire Development Team
 * Copyright (©) 1992 Frank Tore Johansen
 *
 * Deliantra is free software: you can redistribute it and/or modify it under
 * the terms of the Affero GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the Affero GNU General Public License
 * and the GNU General Public License along with this program. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * The authors can be reached via e-mail to <support@deliantra.net>
 */

#include <global.h>
#include <rmg.h>
#include <sproto.h>
#include <rproto.h>

/*  find a character in the maze.  fx and fy are pointers to
    where to find the char.  fx,fy = -1 if not found. */
static void
find_in_layout (int mode, char target, int &fx, int &fy, layout &maze)
{
  int M;
  int i, j;

  fx = -1;
  fy = -1;

  /* if a starting point isn't given, pick one */
  if (mode < 1 || mode > 4)
    M = rmg_rndm (4) + 1;
  else
    M = mode;

  /* four different search starting points and methods so that
     we can do something different for symmetrical maps instead of
     the same damned thing every time. */
  switch (M)
    {
      case 1:
        /* search from top left down/right */
        for (i = 1; i < maze.w; i++)
          for (j = 1; j < maze.h; j++)
            {
              if (maze[i][j] == target)
                {
                  fx = i;
                  fy = j;
                  return;
                }
            }
        break;
        
      case 2:
        /* Search from top right down/left */
        for (i = maze.w - 2; i > 0; i--)
          for (j = 1; j < maze.h - 1; j++)
            {
              if (maze[i][j] == target)
                {
                  fx = i;
                  fy = j;
                  return;
                }
            }
        break;
        
      case 3:
        /* search from bottom-right up-left */
        for (i = maze.w - 2; i > 0; i--)
          for (j = maze.h - 2; j > 0; j--)
            {
              if (maze[i][j] == target)
                {
                  fx = i;
                  fy = j;
                  return;
                }
            }
        break;

      case 4:
        /* search from bottom-left up-right */
        for (i = 1; i < maze.w - 1; i++)
          for (j = maze.h - 2; j > 0; j--)
            {
              if (maze[i][j] == target)
                {
                  fx = i;
                  fy = j;
                  return;
                }
            }
        break;
    }
}

point
layout::find (char target, int mode)
{
  int x, y;

  find_in_layout (mode, target, x, y, *this);

  return point (x, y);
}

/* orientation:  0 means random,
  1 means descending dungeon
  2 means ascending dungeon
  3 means rightward
  4 means leftward
  5 means northward
  6 means southward
*/
void
place_exits (maptile *map, layout &maze, const char *exitstyle, int orientation, random_map_params *RP)
{
  maptile *style_map_down = 0;  /* harder maze */
  maptile *style_map_up = 0;    /* easier maze */
  object *the_exit_down;        /* harder maze */
  object *the_exit_up;          /* easier maze */
  int cx = -1, cy = -1;         /* location of a map center */
  int upx = -1, upy = -1;       /* location of up exit */
  int downx = -1, downy = -1;
  int final_map_exit = 1;

  if (const char *eofm = RP->get_str ("exit_on_final_map", 0))
    if (strstr (eofm, "no"))
      final_map_exit = 0;

  if (!orientation)
    orientation = rmg_rndm (6) + 1;

  switch (orientation)
    {
      case 1:
        {
          style_map_up   = find_style ("/styles/exitstyles/up"  , exitstyle, RP->difficulty);
          style_map_down = find_style ("/styles/exitstyles/down", exitstyle, RP->difficulty);
          break;
        }

      case 2:
        {
          style_map_up   = find_style ("/styles/exitstyles/down", exitstyle, RP->difficulty);
          style_map_down = find_style ("/styles/exitstyles/up"  , exitstyle, RP->difficulty);
          break;
        }

      default:
        {
          style_map_up   =
          style_map_down = find_style ("/styles/exitstyles/generic", exitstyle, RP->difficulty);
          break;
        }
    }

  the_exit_up = style_map_up
    ? style_map_up->pick_random_object (rmg_rndm)->clone ()
    : archetype::get (shstr_exit);

  const char *final_map = RP->get_str ("final_map", 0);

  /* we need a down exit only if we're recursing. */
  if (RP->dungeon_level < RP->dungeon_depth || final_map)
    the_exit_down = style_map_down
      ? style_map_down->pick_random_object (rmg_rndm)->clone ()
      : archetype::get (shstr_exit);
  else
    the_exit_down = 0;

  /* set up the up exit */
  the_exit_up->stats.hp = RP->get_iv  ("origin_x", -1);
  the_exit_up->stats.sp = RP->get_iv  ("origin_y", -1);
  the_exit_up->slaying  = RP->get_str ("origin_map", 0);

  /* figure out where to put the entrance */
  /* begin a logical block */
  {
    /* First, look for a '<' char */
    find_in_layout (0, '<', upx, upy, maze);

    /* next, look for a C, the map center.  */
    find_in_layout (0, 'C', cx, cy, maze);

    /* if we didn't find an up, find an empty place far from the center */
    if (upx == -1 && cx != -1)
      {
        if (cx > RP->Xsize / 2)
          upx = 1;
        else
          upx = RP->Xsize - 2;

        if (cy > RP->Ysize / 2)
          upy = 1;
        else
          upy = RP->Ysize - 2;

        /* find an empty place far from the center */
        if      (upx == 1 && upy == 1) find_in_layout (1, 0, upx, upy, maze);
        else if (upx >  1 && upy == 1) find_in_layout (2, 0, upx, upy, maze);
        else if (upx >  1 && upy  > 1) find_in_layout (3, 0, upx, upy, maze);
        else if (upx == 1 && upy  > 1) find_in_layout (4, 0, upx, upy, maze);
      }

    /* no indication of where to place the exit, so just place it. */
    if (upx == -1)
      find_in_layout (0, 0, upx, upy, maze);

    the_exit_up->x = upx;
    the_exit_up->y = upy;

    /* surround the exits with notices that this is a random map. */
    for (int j = 1; j < 9; j++)
      if (!wall_blocked (map, the_exit_up->x + freearr_x[j], the_exit_up->y + freearr_y[j]))
        {
          object *random_sign = archetype::get (shstr_sign);
          random_sign->msg = format ("This is a random map.\nLevel: %d of %d.\n", RP->dungeon_level - 1, RP->dungeon_depth);
          map->insert (random_sign, the_exit_up->x + freearr_x[j], the_exit_up->y + freearr_y[j], 0, 0);
        }

    /* Block the exit so things don't get dumped on top of it. */
    the_exit_up->move_block = MOVE_ALL;

    insert_ob_in_map (the_exit_up, map, NULL, 0);
    maze[the_exit_up->x][the_exit_up->y] = '<';

    /* set the starting x,y for this map */
    map->enter_x = the_exit_up->x;
    map->enter_y = the_exit_up->y;

    /* first, look for a '>' character */
    find_in_layout (0, '>', downx, downy, maze);

    /* if no > is found use C */
    if (downx == -1)
      {
        downx = cx;
        downy = cy;
      };

    /* make the other exit far away from this one if
       there's no center. */
    if (downx == -1)
      {
        if (upx > RP->Xsize / 2)
          downx = 1;
        else
          downx = RP->Xsize - 2;

        if (upy > RP->Ysize / 2)
          downy = 1;
        else
          downy = RP->Ysize - 2;

        /* find an empty place far from the entrance */
        if      (downx == 1 && downy == 1) find_in_layout (1, 0, downx, downy, maze);
        else if (downx >  1 && downy == 1) find_in_layout (2, 0, downx, downy, maze);
        else if (downx >  1 && downy >  1) find_in_layout (3, 0, downx, downy, maze);
        else if (downx == 1 && downy >  1) find_in_layout (4, 0, downx, downy, maze);
      }

    /* no indication of where to place the down exit, so just place it */
    if (downx == -1)
      find_in_layout (0, 0, downx, downy, maze);

    if (the_exit_down)
      {
        int i = rmg_find_free_spot (the_exit_down, map, downx, downy, 1, SIZEOFFREE1 + 1);

        the_exit_down->x = downx + freearr_x[i];
        the_exit_down->y = downy + freearr_y[i];

        RP->set ("origin_x", (IV)the_exit_down->x);
        RP->set ("origin_y", (IV)the_exit_down->y);

        the_exit_down->msg     = RP->as_shstr ();
        the_exit_down->slaying = shstr_random_map_exit;

        /* the identifier for making a random map. */
        if (RP->dungeon_level >= RP->dungeon_depth && final_map)
          {
            the_exit_down->msg     = 0;
            the_exit_down->slaying = final_map;

            if (final_map_exit)
              if (maptile *new_map = maptile::find_sync (final_map))
                {
                  object *the_exit_back = the_exit_up->arch->instance ();

                  new_map->load_sync ();

                  for (object *tmp = new_map->at (new_map->enter_x, new_map->enter_y).bot; tmp; tmp = tmp->above)
                    /* Remove exit back to previous random map.  There should only be one
                     * which is why we break out.  To try to process more than one
                     * would require keeping a 'next' pointer, ad free_object kills tmp, which
                     * breaks the for loop.
                     */
                    if (tmp->type == EXIT && EXIT_PATH (tmp).starts_with ("?random/"))
                      {
                        tmp->destroy ();
                        break;
                      }

                  /* setup the exit back */
                  the_exit_back->slaying = map->path;
                  the_exit_back->stats.hp = the_exit_down->x;
                  the_exit_back->stats.sp = the_exit_down->y;
                  the_exit_back->x = new_map->enter_x;
                  the_exit_back->y = new_map->enter_y;

                  insert_ob_in_map (the_exit_back, new_map, NULL, 0);
                }
          }

        /* Block the exit so things don't get dumped on top of it. */
        the_exit_down->move_block = MOVE_ALL;
        insert_ob_in_map (the_exit_down, map, NULL, 0);
        maze[the_exit_down->x][the_exit_down->y] = '>';
      }
  }
}

/* this function unblocks the exits.  We blocked them to
   keep things from being dumped on them during the other
   phases of random map generation. */
void
unblock_exits (maptile *map, layout &maze)
{
  int i = 0, j = 0;
  object *walk;

  for (i = 0; i < maze.w; i++)
    for (j = 0; j < maze.h; j++)
      if (maze[i][j] == '>' || maze[i][j] == '<')
        {
          for (walk = GET_MAP_OB (map, i, j); walk != NULL; walk = walk->above)
            {
              if (walk->move_block == MOVE_ALL && walk->type != LOCKED_DOOR)
                {
                  walk->move_block = 0;
                  update_object (walk, UP_OBJ_CHANGE);
                }
            }
        }
}

